Copyright (c) 2007-2008 dhpoware. All Rights Reserved.

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.

-------------------------------------------------------------------------------

Tangent space normal mapping shader for a single spot light source.

We borrow the Direct3D convention of an inner and an outer cone for the spot
light. A direct per-fragment implementation of the OpenGL spot light model will
produce a very sharp falloff around the spotCutoff. Using the inner and outer
cone convention will produce a more softer falloff around the spotCutoff.

To enable the inner and outer cone version of the spot light fragment shader
define the preprocessor symbol USE_INNER_AND_OUTER_CONES. If this symbol is not
defined the fragment shader will use the OpenGL spot light model.

The tangent vector is passed to the vertex shader in gl_MultiTexCoord1. The
tangent vector is assumed to be a four component vector. The tangent vector's
w component indicates the handedness of the local tangent space at this vertex.
The handedness is used to calculate the bitangent vector. The reason for the
inclusion of the handedness component is to allow for triangles with mirrored
texture mappings.

The vertex shader transforms the lightDir, spotDir, viewDir vectors into
tangent space using the rotation matrix:

                  [ tangent.x    tangent.y    tangent.z   ]
    tbnMatrix  =  [ bitangent.x  bitangent.y  bitangent.z ]
                  [ normal.x     normal.y     normal.z    ]

Since GLSL matrices are column-major the tbnMatrix is declared as

    mat3 tbnMatrix = mat3(tangent.x, bitangent.x, normal.x,
                          tangent.y, bitangent.y, normal.y,
                          tangent.z, bitangent.z, normal.z);

and using this matrix we can transform vectors into tangent space as follows:
    
    vec3 lightDir = tbnMatrix * gl_LightSource[0].position.xyz;
    
The fragment shader performs standard Blinn-Phong per fragment lighting. The
only thing different is we read the normal from the normal map texture rather
than using an interpolated vertex normal. Prior to being stored in the normal
map texture the normals are in the range [-1,1]. Once stored in the normal map
texture the normals will be compressed to the range [0,255]. When the normals
are accessed from the fragment shader GLSL will remap then to the range [0,1].
To convert the normals back to the original [-1,1] range we need to perform a
scale and a bias. We do this by:
    
    vec3 n = texture2D(normalMap, gl_TexCoord[0].st).xyz * 2.0 - 1.0;

We don't have to worry about transforming any vectors into tangent space in the
fragment shader because we have already done this in the vertex shader.

The attenuation is calculated using this formula:

	attenuation factor = 1 - ((d * d) / (r * r))

Where
	d = distance between the light's position and the vertex
	r = radius of the light
    
This formula simplifies to:

	lightDir = (gl_LightSource[0].position.xyz - position) / lightRadius
	attenuation factor = 1 - dot(lightDir, lightDir)

Finally the attenuation factor is clamped to prevent negative values:

	attenuation factor = max(0, 1 - dot(lightDir, lightDir))

-------------------------------------------------------------------------------

[vert]

#version 110

uniform float lightRadius;

varying vec3 lightDir;
varying vec3 spotDir;
varying vec3 viewDir;

void main()
{
    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
    gl_TexCoord[0] = gl_MultiTexCoord0;

    vec3 vertexPos = vec3(gl_ModelViewMatrix * gl_Vertex);

    vec3 n = normalize(gl_NormalMatrix * gl_Normal);
    vec3 t = normalize(gl_NormalMatrix * gl_MultiTexCoord1.xyz);
    vec3 b = cross(n, t) * gl_MultiTexCoord1.w;
    
    mat3 tbnMatrix = mat3(t.x, b.x, n.x,
                          t.y, b.y, n.y,
                          t.z, b.z, n.z);
    
    lightDir = (gl_LightSource[0].position.xyz - vertexPos) / lightRadius;
    lightDir = tbnMatrix * lightDir;

    spotDir = gl_LightSource[0].spotDirection;
    spotDir = tbnMatrix * spotDir;

    viewDir = -vertexPos;
    viewDir = tbnMatrix * viewDir;
}

[frag]

#version 110

#define USE_INNER_AND_OUTER_CONES

uniform sampler2D colorMap;
uniform sampler2D normalMap;

#if defined(USE_INNER_AND_OUTER_CONES)
    uniform float cosOuterCone;
    uniform float cosInnerCone;
#endif

varying vec3 lightDir;
varying vec3 spotDir;
varying vec3 viewDir;

void main()
{
	vec3 l = lightDir;
    float atten = max(0.0, 1.0 - dot(l, l));
    
    l = normalize(l);

    float spotDot = dot(-l, normalize(spotDir));

#if defined(USE_INNER_AND_OUTER_CONES)
    float spotEffect = smoothstep(cosOuterCone, cosInnerCone, spotDot);
#else
    float spotEffect = (spotDot < gl_LightSource[0].spotCosCutoff)
                        ? 0.0 : pow(spotDot, gl_LightSource[0].spotExponent);
#endif

    atten *= spotEffect;

    vec3 n = normalize(texture2D(normalMap, gl_TexCoord[0].st).xyz * 2.0 - 1.0);
    vec3 v = normalize(viewDir);
    vec3 h = normalize(l + v);

    float nDotL = max(0.0, dot(n, l));
    float nDotH = max(0.0, dot(n, h));
    float power = (nDotL == 0.0) ? 0.0 : pow(nDotH, gl_FrontMaterial.shininess);
    
    vec4 ambient = gl_FrontLightProduct[0].ambient;
    vec4 diffuse = gl_FrontLightProduct[0].diffuse * nDotL * atten;
    vec4 specular = gl_FrontLightProduct[0].specular * power * atten;
    vec4 color = gl_FrontLightModelProduct.sceneColor + ambient + diffuse + specular;
    
    gl_FragColor = color * texture2D(colorMap, gl_TexCoord[0].st);
}
